#include "PrecompiledHeader.h"

#include "CDPlayerThread.h"
#include "../Log.h"
#include "../gui/FormMain.h"
#include <fmod_errors.h>
//**********************************************************************************************************
using namespace JokersPlayer;
extern FormMain * g_formMain;
//**********************************************************************************************************
#define CHECKFMODERROR( X) CDPlayerThread::_checkFMODError( X, __LINE__, __FILE__)
#define SAFE_RELEASE(s) if (s != NULL) {CHECKFMODERROR( s->release()); s = NULL;}
//**********************************************************************************************************

CDPlayerThread :: CDPlayerThread()
	:	PlayerThread(),
		m_mediaTime		( 0.0),
		m_currentTime	( 0.0),
		m_volume		( 1.0)
{
	unsigned int l_version;

	CHECKFMODERROR( FMOD::System_Create( & m_system));
	CHECKFMODERROR( m_system->getVersion( & l_version));

	if (l_version < FMOD_VERSION)
	{
		wxString l_msg = wxT( "Warning! You are using an old version of FMOD [0x");
		l_msg << l_version << wxT( "].  This program requires [0x") << FMOD_VERSION << wxT( "]");
		Log::LogMessage( l_msg);
	}

	FMOD_OUTPUTTYPE l_outputMode[7] =
	{
		FMOD_OUTPUTTYPE_AUTODETECT,
		FMOD_OUTPUTTYPE_DSOUND,
		FMOD_OUTPUTTYPE_WINMM,
		FMOD_OUTPUTTYPE_ASIO,
		FMOD_OUTPUTTYPE_NOSOUND
	};

	for (size_t i = 0 ; i < 5 ; i ++)
	{
		CHECKFMODERROR( m_system->setOutput( l_outputMode[i]));

		if (CHECKFMODERROR( m_system->init( 4095, FMOD_INIT_NORMAL, 0)))
		{
			Log::LogMessage( wxT( "FMOD is initialized with Output type : [") + _outputTypeToStr( l_outputMode[i]) + wxT( "]"));
			break;
		}
	}
}

//**********************************************************************************************************

void * CDPlayerThread :: Entry()
{
	Log::LogMessage( "AudioPlayerThread :: Execute - Begin");
	unsigned int l_position;
	double l_barPosition;
	bool l_isPlaying = true;
	while (m_status != STATUS_STOP && IsAlive() && IsRunning() && ! TestDestroy() && l_isPlaying)
	{
		CHECKFMODERROR( m_channel->isPlaying( & l_isPlaying));
		CHECKFMODERROR( m_channel->getPosition( & l_position, FMOD_TIMEUNIT_MS));
		l_barPosition = l_position * 1.0 / m_mediaTime;
		g_formMain->m_mediaBar->SetValue( static_cast <int>( l_barPosition * 1000));
	}
	_stopPlayback();
	Cleanup();
	g_formMain->EndOfMedia();
	Log::LogMessage( "AudioPlayerThread :: Execute - End");

	return NULL;
}

//**********************************************************************************************************

bool CDPlayerThread :: LoadMedia( const wxString & p_url)
{
	m_drive = p_url.substr( 0, 2);
	wxString l_trackStr = p_url.substr( 2, wxString::npos);
	long l_track;
	l_trackStr.ToLong( & l_track);
	m_trackNumber = (int)l_track;
	FMOD_MODE l_mode = FMOD_DEFAULT;
	CHECKFMODERROR( m_system->createStream( m_drive.char_str(), l_mode, NULL, & m_sound));
	CHECKFMODERROR( m_sound->getNumSubSounds( & m_tracksNumber));
	Log::LogMessage( "Number of tracks : %i/%i", m_trackNumber, m_tracksNumber);
	if (m_trackNumber < m_tracksNumber)
	{
		CHECKFMODERROR( m_sound->getSubSound( m_trackNumber, & m_track));
		if ( ! CHECKFMODERROR( m_system->playSound( FMOD_CHANNEL_FREE, m_track, true, & m_channel)))
		{
			Log::LogMessage( "Play - Can't load track %i", m_trackNumber);
			return false;
		}
		CHECKFMODERROR( m_track->getLength( & m_mediaTime, FMOD_TIMEUNIT_MS));
		m_status = STATUS_READY;
		return true;
	}
	return false;
}

//**********************************************************************************************************

void CDPlayerThread :: StartPlayback()
{
	if (m_status < STATUS_READY)
	{
		return;
	}

	if (IsPaused())
	{
		Resume();
	}

	if ( ! CHECKFMODERROR( m_channel->setPaused( false)))
	{
		Log::LogMessage( "Play - Can't unpause");
		return;
	}

	m_status = STATUS_PLAY;
}

//**********************************************************************************************************

void CDPlayerThread :: PausePlayback()
{
	if (m_status == STATUS_PLAY)
	{
		return;
	}

	if ( ! CHECKFMODERROR( m_channel->setPaused( true)))
	{
		Log::LogMessage( "Play - Can't pause");
		return;
	}

	m_status = STATUS_PAUSE;

	if ( ! IsPaused())
	{         
		Pause();
	}
}

//**********************************************************************************************************

void CDPlayerThread :: Cleanup()
{
	if (m_status == STATUS_PLAY || m_status == STATUS_PAUSE)
	{
		CHECKFMODERROR( m_channel->stop());
	}
	SAFE_RELEASE( m_track);
	SAFE_RELEASE( m_sound);
	CHECKFMODERROR( m_system->close());
	SAFE_RELEASE( m_system);
}

//**********************************************************************************************************

void CDPlayerThread :: SetPosition( float p_position)
{
	unsigned int l_position = static_cast <unsigned int>( p_position * m_mediaTime);
	CHECKFMODERROR( m_channel->setPosition( l_position, FMOD_TIMEUNIT_MS));
}

//**********************************************************************************************************

wxString CDPlayerThread :: GetMediaTime()
{
	wxString l_str;
	unsigned int l_nbMin = m_mediaTime / 60000;
	unsigned int l_nbSec = (m_mediaTime % 60000) / 1000;
	l_str << l_nbMin << wxT( "min") << l_nbSec << wxT( "s");
	Log::LogMessage( "GetMediaTime : %s", static_cast <char *>( l_str.char_str()));
	return l_str;
}

//**********************************************************************************************************

bool CDPlayerThread :: SetAudio(int volume, int balance)
{
	return CHECKFMODERROR( m_channel->setVolume( m_volume));
}

//**********************************************************************************************************

void CDPlayerThread :: Mute( bool p_mute)
{
	if ( ! CHECKFMODERROR( m_channel->setMute( p_mute)))
	{
		Log::LogMessage( "Play - Can't mute");
		return;
	}
	m_muted = p_mute;
}

//**********************************************************************************************************

void CDPlayerThread :: FastForward()
{
	if (m_status == STATUS_PLAY || m_status == STATUS_PAUSE)
	{           
		unsigned int l_position;
		CHECKFMODERROR( m_channel->getPosition( & l_position, FMOD_TIMEUNIT_MS));
		l_position += 1000;
		if (l_position > m_mediaTime)
		{
			l_position = m_mediaTime;
		}
		CHECKFMODERROR( m_channel->setPosition( l_position, FMOD_TIMEUNIT_MS));
	}
}

//**********************************************************************************************************

void CDPlayerThread :: FastBack()
{
	if (m_status == STATUS_PLAY || m_status == STATUS_PAUSE)
	{           
		unsigned int l_position;
		CHECKFMODERROR( m_channel->getPosition( & l_position, FMOD_TIMEUNIT_MS));
		l_position -= 1000;
		if (l_position < 0)
		{
			l_position = 0;
		}
		CHECKFMODERROR( m_channel->setPosition( l_position, FMOD_TIMEUNIT_MS));
	}
}

//**********************************************************************************************************

void CDPlayerThread :: _stopPlayback()
{
	if (m_status != STATUS_PLAY && m_status != STATUS_PAUSE)
	{
		return;
	}

	if ( ! CHECKFMODERROR( m_channel->stop()))
	{
		Log::LogMessage( "Play - Can't stop");
		return;
	}

	m_status = STATUS_STOP;
	m_currentTime = 0;
}

//**********************************************************************************************************

bool CDPlayerThread :: _checkFMODError( FMOD_RESULT p_result, unsigned int p_line, const char * p_file)
{
	if (p_result == FMOD_OK)
	{
		return true;
	}

	char * l_errorString = const_cast <char *>( FMOD_ErrorString( p_result));
	wxString l_msg = wxT( "FMOD error (");
	l_msg << p_result;
	l_msg += wxString( wxT( ") : [ ") + wxString( l_errorString, wxConvLibc) + wxT( "] in \"") + wxString( p_file, wxConvLibc) + wxT( "\"@"));
	l_msg << p_line;

	Log::LogMessage( l_msg);
	return false;
}

//**********************************************************************************************************

const wxString CDPlayerThread :: _outputTypeToStr( FMOD_OUTPUTTYPE p_outputType)
{
	switch (p_outputType)
	{
		case FMOD_OUTPUTTYPE_AUTODETECT:	return wxT( "Autodetect");
		case FMOD_OUTPUTTYPE_NOSOUND:		return wxT( "No sound");
		case FMOD_OUTPUTTYPE_DSOUND:		return wxT( "DirectSound");
		case FMOD_OUTPUTTYPE_WINMM:			return wxT( "Windows Multimedia");
		case FMOD_OUTPUTTYPE_ASIO:			return wxT( "ASIO");
		case FMOD_OUTPUTTYPE_OSS:			return wxT( "OSS");
		case FMOD_OUTPUTTYPE_ALSA:			return wxT( "ALSA");
		case FMOD_OUTPUTTYPE_ESD:			return wxT( "ESD");
		case FMOD_OUTPUTTYPE_SOUNDMANAGER:	return wxT( "Sound Manager");
		case FMOD_OUTPUTTYPE_COREAUDIO:		return wxT( "Core Audio");
		case FMOD_OUTPUTTYPE_WAVWRITER:		return wxT( "Wave writer");
		case FMOD_OUTPUTTYPE_NOSOUND_NRT:	return wxT( "Non real-time no sound");
		case FMOD_OUTPUTTYPE_WAVWRITER_NRT:	return wxT( "Non real-time Wave writer");
		case FMOD_OUTPUTTYPE_XBOX:			return wxT( "XBox");
		case FMOD_OUTPUTTYPE_XBOX360:		return wxT( "XBox 360");
		case FMOD_OUTPUTTYPE_PS2:			return wxT( "Playstation 2");
		case FMOD_OUTPUTTYPE_PS3:			return wxT( "Playstation 3");
		case FMOD_OUTPUTTYPE_GC:			return wxT( "GameCube");
		case FMOD_OUTPUTTYPE_WII:			return wxT( "Wii");
		case FMOD_OUTPUTTYPE_PSP:			return wxT( "PSP");
		case FMOD_OUTPUTTYPE_UNKNOWN:		return wxT( "Unknown");
#ifndef WIN32
		case FMOD_OUTPUTTYPE_OPENAL:		return wxT( "OpenAL");
		case FMOD_OUTPUTTYPE_WASAPI:		return wxT( "Wasapi");
#endif
		case FMOD_OUTPUTTYPE_MAX:			break;
		case FMOD_OUTPUTTYPE_FORCEINT:		break;
	}

	wxString l_str = wxT( "[");
	l_str << p_outputType << wxT( "] isn't a valid output type...");
	return l_str;
}

//**********************************************************************************************************
